//Module to implment drag & dropping of files and text in the edit controls


#include "StdAfx.h"
#include "mydroptarget.h"


CMyDropTarget::CMyDropTarget(CEdit* pDropCtrl)
{
	//Constructor
	ASSERT(pDropCtrl);

	m_lRefCount = 1;
	m_fAllowDrop = FALSE;
	pDropWnd = pDropCtrl;

	pEditWnd = pDropCtrl;
}

CMyDropTarget::~CMyDropTarget(void)
{
	m_Edit.Detach();
}

HRESULT __stdcall CMyDropTarget::QueryInterface (REFIID iid, void** ppvObject)
{
	//IUnknown::QueryInterface
	if(iid == IID_IDropTarget || iid == IID_IUnknown)
	{
		AddRef();
		*ppvObject = this;
		return S_OK;
	}
	else
	{
		*ppvObject = 0;
		return E_NOINTERFACE;
	}
}

ULONG __stdcall CMyDropTarget::AddRef(void)
{
	//IUnknown::AddRef
	return InterlockedIncrement(&m_lRefCount);
}	

ULONG __stdcall CMyDropTarget::Release(void)
{
	//IUnknown::Release
	LONG count = InterlockedDecrement(&m_lRefCount);
		
	if(count == 0)
	{
		delete this;
		return 0;
	}
	else
	{
		return count;
	}
}

HRESULT __stdcall CMyDropTarget::DragEnter(IDataObject* pDataObject, DWORD grfKeyState, POINTL pt, DWORD* pdwEffect)
{
	//IDropTarget::DragEnter
	m_fAllowDrop = QueryDataObject(pDataObject);
	
	if(m_fAllowDrop)
	{
		// get the dropeffect based on keyboard state
		*pdwEffect = DropEffect(grfKeyState, pt, *pdwEffect);

		pEditWnd->SetFocus();

		//Select all data in control
		SelectControl(pt);
	}
	else
	{
		*pdwEffect = DROPEFFECT_NONE;
	}

	return S_OK;
}

HRESULT __stdcall CMyDropTarget::DragOver(DWORD grfKeyState, POINTL pt, DWORD* pdwEffect)
{
	//IDropTarget::DragOver
	if(m_fAllowDrop)
	{
		*pdwEffect = DropEffect(grfKeyState, pt, *pdwEffect);

		//Select all data in control
		SelectControl(pt);
	}
	else
	{
		*pdwEffect = DROPEFFECT_NONE;
	}

	return S_OK;
}

HRESULT __stdcall CMyDropTarget::DragLeave(void)
{
	//IDropTarget::DragLeave

	return S_OK;
}

HRESULT __stdcall CMyDropTarget::Drop(IDataObject* pDataObject, DWORD grfKeyState, POINTL pt, DWORD* pdwEffect)
{
	//IDropTarget::Drop

	//Select all data in control
	SelectControl(pt);

	if(m_fAllowDrop)
	{
		//Data dropped
		DropData(pDataObject);

		*pdwEffect = DropEffect(grfKeyState, pt, *pdwEffect);
	}
	else
	{
		*pdwEffect = DROPEFFECT_NONE;
	}

	return S_OK;
}

BOOL CMyDropTarget::QueryDataObject(IDataObject* pDataObject)
{
	//Determines if 'pDataObject' object is Text
	//RETURN: = TRUE if data type is supported

	//'pnOutDataType' is filled out (if returns TRUE):
	//		= 1 - Link;
	//		= 2 - Text
	// http://msdn.microsoft.com/en-us/library/windows/desktop/ff729168(v=vs.85).aspx
	static FORMATETC fmtetc_txt = {CF_UNICODETEXT, 0, DVASPECT_CONTENT, -1, TYMED_HGLOBAL};

	BOOL bRes = FALSE;
	int nType = 0;
	HRESULT hr;

	//See if our window is enabled
	if(!pEditWnd->IsWindowEnabled())
		return FALSE;


	//IEnumFORMATETC *pEnum = NULL;
	//hr = pDataObject->EnumFormatEtc(DATADIR_GET, &pEnum);
	//if(pEnum)
	//{
	//	FORMATETC fmt = {0};
	//	ULONG nNumRet = 0;
	//	for(;;)
	//	{
	//		hr = pEnum->Next(1, &fmt, &nNumRet);
	//		if(FAILED(hr) ||
	//			nNumRet == 0)
	//		{
	//			break;
	//		}
	//	}

	//	pEnum->Release();
	//}


	hr = pDataObject->QueryGetData(&fmtetc_txt);
	if(hr == S_OK)
	{
		bRes = TRUE;
		nType = 2;
	}


	return bRes;
}

DWORD CMyDropTarget::DropEffect(DWORD grfKeyState, POINTL pt, DWORD dwAllowed)
{
	//Determines type of operation for dragging
	DWORD dwEffect = 0;

	//Determine type
	DWORD nDrgTypeMove = DROPEFFECT_COPY;//DROPEFFECT_MOVE;
	DWORD nDrgTypeCopy = DROPEFFECT_COPY;

	//See if any keys pressed
	if(grfKeyState & MK_CONTROL)
	{
		dwEffect = dwAllowed & nDrgTypeCopy;
	}
	else if(grfKeyState & MK_SHIFT)
	{
		dwEffect = dwAllowed & nDrgTypeMove;
	}

	//If no effect, use the one from dropsource
	if(dwEffect == 0)
	{
		if(dwAllowed & DROPEFFECT_COPY)
			dwEffect = nDrgTypeCopy;
		if(dwAllowed & DROPEFFECT_MOVE)
			dwEffect = nDrgTypeMove;
		if(dwAllowed & DROPEFFECT_LINK)
			dwEffect = DROPEFFECT_LINK;
	}
	
	return dwEffect;
}

void CMyDropTarget::SelectControl(POINTL pt)
{
	//Select edit control when mouse it over it
	ASSERT(pEditWnd);

	////Get the character position of mouse
	//DWORD curpos; 
	//pEditWnd->ScreenToClient((POINT*)&pt);
	//curpos = (DWORD)pEditWnd->SendMessage(EM_CHARFROMPOS, 0, MAKELPARAM(pt.x, pt.y));

	//// set cursor position
	//pEditWnd->SendMessage(EM_SETSEL, LOWORD(curpos), LOWORD(curpos));

	pEditWnd->SetSel(0, -1);
}

BOOL CMyDropTarget::RegisterDragDropWnd(CEdit* pEditCtrl, IDropTarget **ppDropTarget)
{
	//Register edit control for Drag&Drop file operation
	//(IN ANY CASE UnregisterDragDropWnd() SHOULD BE CALLED WHEN NOT NEEDED!)
	//RETURN: = TRUE if done
	ASSERT(pEditCtrl);
	ASSERT(ppDropTarget);

	//Create instance of the class/interface
	CMyDropTarget *pDropTarget = new CMyDropTarget(pEditCtrl);
	*ppDropTarget = pDropTarget;

	//See if we were able to find edit control in the combo box
	if(!pDropTarget->pEditWnd)
		return FALSE;

	//Acquire a strong lock
	if(CoLockObjectExternal(pDropTarget, TRUE, FALSE) == S_OK)
	{
		//Now register control
		if(RegisterDragDrop(pEditCtrl->GetSafeHwnd(), pDropTarget) == S_OK)
		{
			//All done!
			return TRUE;
		}
	}

	return FALSE;
}

BOOL CMyDropTarget::UnregisterDragDropWnd(CEdit* pEditCtrl, IDropTarget *pDropTarget)
{
	//Unregister drag & drop
	if(pEditCtrl)
	{
		//Revoke drag & drop
		RevokeDragDrop(pEditCtrl->GetSafeHwnd());
		pEditCtrl = NULL;
	}

	if(pDropTarget)
	{
		//Remove the strong lock
		VERIFY(CoLockObjectExternal(pDropTarget, FALSE, TRUE) == S_OK);

		//Release our own reference
		pDropTarget->Release();

		pDropTarget = NULL;
	}

	return TRUE;
}

BOOL CMyDropTarget::DropData(IDataObject* pDataObject)
{
	//Called when pDataObject object is dropped into the edit control
	//RETURN: = TRUE if done
	ASSERT(pDataObject);
	ASSERT(pEditWnd);

	BOOL bRes = FALSE;
	STGMEDIUM stgmed;
	CString text;

	//See if it's text first
	static FORMATETC fmtetc_txt = {CF_UNICODETEXT, 0, DVASPECT_CONTENT, -1, TYMED_HGLOBAL};
	if(pDataObject->QueryGetData(&fmtetc_txt) == S_OK)
	{
		//We got text, now obtain it as HGLOBAL
		if(pDataObject->GetData(&fmtetc_txt, &stgmed) == S_OK &&
			stgmed.tymed == TYMED_HGLOBAL)
		{
			//Lock the HGLOBAL block
			PVOID pData = GlobalLock(stgmed.hGlobal);
			if(pData)
			{
				//Get text
				bRes = TRUE;
				text = (TCHAR*)pData;

				//Unlock memory
				GlobalUnlock(stgmed.hGlobal);
			}

			//Release the data using the COM API
			ReleaseStgMedium(&stgmed);
		}
	}


	//See if we got anything
	if(bRes)
	{
		//Read selection
		DWORD dwCurSel = pEditWnd->GetSel();

		//Set data into edit control
		CString str = text;

		//Set text back into control
		int nLimit = pEditWnd->GetLimitText();
		if(str.GetLength() > nLimit)
		{
			//Have to shorted the string
			str = str.Left(nLimit);
		}

		//Set text now
		pEditWnd->SetWindowText(str);

		//Set selection back
		pEditWnd->SetSel(dwCurSel);
	}

	return bRes;
}
